(*Author: Giuliano Losa *)

header {* The ALM Automata specification*}
theory ALM
imports IOA LCP
begin

typedecl client
  -- {*A non-empty set of clients*}
typedecl data
  -- {*Data contained in requests*}
datatype request = 
  -- {*A request is composed of a sender and data*}
  Req client data

fun request_snd :: "request \<Rightarrow> client"
  where "request_snd (Req c _) = c"

type_synonym hist = "request list" 
  -- {*Type of histories of requests.*}

datatype ALM_action = 
  -- {*The actions of the ALM automaton*}
  Invoke client "request" 
  | Commit client nat "hist"
  | Switch client nat "hist" "request"
  | Initialize nat "hist"
  | Linearize nat "hist"
  | Abort nat

datatype phase = Sleep | Pending | Ready | Aborted
  -- {*Executions phases of a client*}

definition linearizations :: "request set \<Rightarrow> hist set"
  -- {*The possible linearizations of a set of requests*}
  where 
  "linearizations reqs \<equiv> {h . set h \<subseteq> reqs \<and> distinct h}"

definition postfix_all :: "hist \<Rightarrow> hist set \<Rightarrow> hist set"
  -- {*appends to the right the first argument to every member of the history set*}
  where 
  "postfix_all h hs \<equiv> {h' . \<exists> h'' . h' = h'' @ h \<and> h'' \<in> hs}"

(*
datatype act_class = In | Out | Internal | None

primrec ALM_asig_p :: "nat \<Rightarrow> nat \<Rightarrow> ALM_action \<Rightarrow> act_class"
where
  "ALM_asig_p id1 id2 (Invoke _ _) = In" |
  "ALM_asig_p id1 id2 (Switch _ i _ _) = (if i = id1 \<and> id1 \<noteq> 0 then In else if i = id2 then Out else None)" |
  "ALM_asig_p id1 id2 (Commit _ i _) = (if id1 \<le> i \<and> i < id2 then Out else None)" |
  "ALM_asig_p id1 id2 (Abort i) = (if i = id1 then Internal else None)" |
  "ALM_asig_p id1 id2 (Linearize i _) = (if i = id1 then Internal else None)" |
  "ALM_asig_p id1 id2 (Initialize i _) = (if i = id1 then Internal else None)"

primrec is_input :: "act_class \<Rightarrow> bool"
where
  "is_input In = True" |
  "is_input Out = False" |
  "is_input Internal = False" |
  "is_input None = False"
primrec is_output :: "act_class \<Rightarrow> bool"
where
  "is_output In = False" |
  "is_output Out = True" |
  "is_output Internal = False" |
  "is_output None = False"
primrec is_internal :: "act_class \<Rightarrow> bool"
where
  "is_internal In = False" |
  "is_internal Out = False" |
  "is_internal Internal = True" |
  "is_internal None = False"
  
definition
  ALM_asig :: "nat \<Rightarrow> nat \<Rightarrow> ALM_action signature" 
  -- {*The action signature of ALM automata*}
  -- {*Input actions, output actions, and internal actions*}
  where
  "ALM_asig id1 id2 \<equiv> (
    {act . is_input (ALM_asig_p id1 id2 act)},
    {act . is_output (ALM_asig_p id1 id2 act)},
    {act . is_internal (ALM_asig_p id1 id2 act)} )"
*)

definition
  ALM_asig :: "nat \<Rightarrow> nat \<Rightarrow> ALM_action signature" 
  -- {*The action signature of ALM automata*}
  -- {*Input actions, output actions, and internal actions*}
  where
  "ALM_asig id1 id2 \<equiv> (
    {act . \<exists> c r h . 
             act = Invoke c r | act = Switch c id1 h r},
    {act . \<exists> c h r id' . 
             id1 <= id' \<and> id' < id2 \<and> act = Commit c id' h
             | act = Switch c id2 h r},
    {act . \<exists> h . 
             act = Abort id1
             | act = Linearize id1 h 
             | act = Initialize id1 h} )"

record ALM_state = 
  -- {*The state of the ALM automata *}
  pending :: "client \<Rightarrow> request"
  -- {*Associates a pending request to a client process*}
  initHists :: "hist set"
  -- {*The set of init histories submitted by clients*}
  phase :: "client \<Rightarrow> phase"
  -- {*Associates a phase to a client process*}
  hist :: "hist"
  -- {*Represents the chosen linearization of the concurrent history of the current instance only*}
  aborted :: "bool"
  initialized :: "bool"

lemma ALM_state_ext:
  fixes s1::ALM_state and s2
  shows "(s1 = s2) = (pending s1 = pending s2 \<and> initHists s1 = initHists s2 \<and> phase s1 = phase s2 \<and> hist s1 = hist s2 \<and> aborted s1 = aborted s2 \<and> initialized s1 = initialized s2)"
  using assms by fastforce

definition pendingReqs :: "ALM_state \<Rightarrow> request set" 
  -- {*the set of requests that have been invoked but that are not yet in the hist parameter*}
  where
  "pendingReqs s \<equiv> {r . \<exists> c . 
     r = pending s c
     \<and> r \<notin> set (hist s)
     \<and> phase s c \<in> {Pending, Aborted}}"

definition initValidReqs :: "ALM_state \<Rightarrow> request set" 
  -- {*any request that appears in an init hist after the longest common prefix or that is pending*}
  where
  "initValidReqs s \<equiv> {r . 
     (r \<in> pendingReqs s  \<or> (\<exists> h \<in> initHists s . r \<in> set h))
     \<and> r \<notin> set (l_c_p (initHists s))}"

definition Invoke_trans :: "nat \<Rightarrow> nat \<Rightarrow> client \<Rightarrow> request \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "Invoke_trans id1 id2 c r s s' \<equiv>  
      (if phase s c = Ready \<and> request_snd r = c \<and> r \<notin> set (hist s)
      then s' = s\<lparr>pending := (pending s)(c := r),
                  phase := (phase s)(c := Pending)\<rparr>
      else s' = s)"
      
definition Linearize_trans :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> hist \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "Linearize_trans id1 id2 i h s s' \<equiv>   
      (initialized s \<and> \<not> aborted s
      \<and> h \<in> postfix_all (hist s) (linearizations (pendingReqs s ))
      \<and> s' = s\<lparr>hist := h\<rparr>)"
    
fun awake :: "phase \<Rightarrow> bool"
where "awake Sleep = False" | "awake _ = True"

definition Initialize_trans :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> hist \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "Initialize_trans id1 id2 i h s s' \<equiv>  
      ((\<exists> c . awake (phase s c)) \<and> \<not> aborted s \<and> \<not> initialized s
      \<and> h \<in> postfix_all (l_c_p (initHists s)) (linearizations (initValidReqs s))
      \<and> s' = s\<lparr>hist := h, initialized := True\<rparr>)"
      
definition Commit_trans :: "nat \<Rightarrow> nat \<Rightarrow> client \<Rightarrow> nat \<Rightarrow> hist \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "Commit_trans id1 id2 c i h s s' \<equiv>  
      (phase s c = Pending \<and> pending s c \<in> set (hist s)
      \<and> h = dropWhile (\<lambda> r . r \<noteq> pending s c) (hist s)
      \<and> s' = s \<lparr>phase := (phase s)(c := Ready)\<rparr>)"
        
definition Abort_trans :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "Abort_trans id1 id2 i s s' \<equiv>  
      (\<not> aborted s \<and> (\<exists> c . phase s c \<noteq> Sleep)
      \<and> s' = s\<lparr>aborted:= True\<rparr>)"
      
definition Switch_trans1 :: "nat \<Rightarrow> nat \<Rightarrow> client \<Rightarrow> nat \<Rightarrow> hist \<Rightarrow> request \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
  -- "The case in which i = id1"
where
  "Switch_trans1 id1 id2 c i h r s s' \<equiv>
        (if phase s c = Sleep
         then s' = s \<lparr>initHists := {h} \<union> (initHists s), 
                     phase := (phase s)(c := Pending), 
                     pending := (pending s)(c := r)\<rparr>
         else s' = s)"
      
definition Switch_trans2 :: "nat \<Rightarrow> nat \<Rightarrow> client \<Rightarrow> nat \<Rightarrow> hist \<Rightarrow> request \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
  -- "The case in which i = id2"
where
  "Switch_trans2 id1 id2 c i h r s s' \<equiv>
     aborted s 
     \<and> phase s c = Pending \<and> r = pending s c 
     \<and> (if initialized s
       then (h \<in> postfix_all (hist s) (linearizations (pendingReqs s )))
       else (h \<in> postfix_all (l_c_p (initHists s)) (linearizations (initValidReqs s))))
     \<and> s' = s\<lparr>phase := (phase s)(c := Aborted)\<rparr>"
  
fun ALM_trans_p :: "nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_action \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "ALM_trans_p id1 id2 s (Invoke c r) s' = Invoke_trans id1 id2 c r s s'" |      
  "ALM_trans_p id1 id2 s (Linearize i h) s' = Linearize_trans id1 id2 i h s s'" |      
  "ALM_trans_p id1 id2 s (Initialize i h) s' = Initialize_trans id1 id2 i h s s'" |      
  "ALM_trans_p id1 id2 s (Commit c i h) s' = Commit_trans id1 id2 c i h s s'" | 
  "ALM_trans_p id1 id2 s (Abort i) s' = Abort_trans id1 id2 i s s'" |       
  "ALM_trans_p id1 id2 s (Switch c i h r) s' = 
    ((i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans1 id1 id2 c i h r s s')
    \<or> (i = id2 \<and> Switch_trans2 id1 id2 c i h r s s'))"

definition ALM_trans :: "nat \<Rightarrow> nat \<Rightarrow> (ALM_action, ALM_state)transition set"
where
  "ALM_trans id1 id2 \<equiv> { (s,a,s') . ALM_trans_p id1 id2 s a s'}"

definition ALM_start :: "nat \<Rightarrow> ALM_state set" 
  -- {*the set of start states*}
  where
  "ALM_start i \<equiv> { s . 
    \<forall> c . phase s c = (if i \<noteq> 0 then Sleep else Ready) 
    \<and> hist s = [] 
    \<and> \<not> aborted s 
    \<and> (if i \<noteq> 0 then \<not> initialized s else initialized s)
    \<and> initHists s = {}}"

definition ALM_ioa :: "nat \<Rightarrow> nat \<Rightarrow> (ALM_action, ALM_state)ioa" 
  -- {*The ALM automaton*}
  where
  "ALM_ioa id1 id2 \<equiv> 
     (ALM_asig id1 id2, 
      ALM_start id1, 
      ALM_trans id1 id2,
      {}, {})"

type_synonym compo_state = "ALM_state \<times> ALM_state"

definition composeALMs :: "nat \<Rightarrow> nat \<Rightarrow> (ALM_action, compo_state) ioa" 
  -- {*the composition of two ALMs*}
  where
  "composeALMs id1 id2 \<equiv> 
     hide (ALM_ioa 0 id1 || ALM_ioa id1 id2)
          {act . EX c tr r . act = Switch c id1 tr r}"

end